#!/usr/bin/env pythonw
'''
popup for 2D XRD mask file

'''

import os
import numpy as np

import wx

from wxmplot.imagepanel import ImagePanel
from .ImageControlsFrame import ImageToolboxFrame

from larch.io import tifffile
from larch.utils import get_cwd
# HAS_pyFAI = False
# try:
#     import pyFAI
#     import pyFAI.calibrant
#     from pyFAI.calibration import Calibration
#     HAS_pyFAI = True
# except ImportError:
#     pass

# HAS_fabio = False
# try:
#     import fabio
#     HAS_fabio = True
# except ImportError:
#     pass

###################################

class MaskToolsPopup(wx.Frame):

    def __init__(self,parent):

        self.frame = wx.Frame.__init__(self, parent, title='Create mask',size=(800,600))

        self.parent = parent
        self.statusbar = self.CreateStatusBar(2,wx.CAPTION )

        try:
            self.raw_img = parent.plt_img ## raw_img or flp_img or plt_img mkak 2016.10.28
        except:
            self.loadIMAGE()


        self.setDefaults()


        self.Init()
        self.Show()

#        wx.Window.GetEffectiveMinSize
#        wx.GetBestSize(self)


    def Init(self):

        self.panel = wx.Panel(self)

        self.MainSizer()

        self.framebox = wx.BoxSizer(wx.VERTICAL)
        self.framebox.Add(self.mainbox, flag=wx.ALL|wx.EXPAND, border=10)

        ###########################
        ## Pack all together in self.panel
        self.panel.SetSizer(self.framebox)


    def setDefaults(self):

        self.area_list = []

    def MainSizer(self):

        self.mainbox = wx.BoxSizer(wx.VERTICAL)

        ###########################
        ## -----> Main Panel
        self.hmain = wx.BoxSizer(wx.HORIZONTAL)

        self.ImageSizer()
        self.DrawNewSizer()

        self.hmain.Add(self.imagebox,proportion=1,flag=wx.ALL|wx.EXPAND, border=10)
        self.hmain.Add(self.toolbox, flag=wx.ALL, border=10)

        self.mainbox.Add(self.hmain, flag=wx.ALL|wx.EXPAND, border=10)


    def DrawNewSizer(self):

        self.toolbox = wx.BoxSizer(wx.VERTICAL)

        ###########################
        ## Directions
        nwbx = wx.StaticBox(self.panel,label='Drawing Tools', size=(100, 50))
        drawbox = wx.StaticBoxSizer(nwbx,wx.VERTICAL)

        ###########################
        ## Drawing tools
        hbox_shp = wx.BoxSizer(wx.HORIZONTAL)
        self.txt_shp = wx.StaticText(self.panel, label='DRAWING SHAPE')
        shapes = ['circle','pixel','polygon','square']

        self.ch_shp = wx.Choice(self.panel,choices=shapes)

        self.ch_shp.Bind(wx.EVT_CHOICE,self.ShapeChoice)

        hbox_shp.Add(self.txt_shp, flag=wx.RIGHT, border=8)
        hbox_shp.Add(self.ch_shp, flag=wx.EXPAND, border=8)

        drawbox.Add(hbox_shp, flag=wx.ALL|wx.EXPAND, border=10)

        ###########################
        ## Drawn Areas
        vbox_areas = wx.BoxSizer(wx.VERTICAL)
        hbox_areas = wx.BoxSizer(wx.HORIZONTAL)

        self.slct_area = wx.ListBox(self.panel, 26, wx.DefaultPosition, (170, 130), self.area_list, wx.LB_SINGLE)

        self.btn_SHWarea = wx.Button(self.panel,label='SHOW')
        self.btn_DELarea = wx.Button(self.panel,label='DELETE')

        self.btn_SHWarea.Bind(wx.EVT_BUTTON,self.showAREA)
        self.btn_DELarea.Bind(wx.EVT_BUTTON,self.deleteAREA)

        hbox_areas.Add(self.btn_SHWarea, flag=wx.BOTTOM, border=10)
        hbox_areas.Add(self.btn_DELarea, flag=wx.BOTTOM, border=10)

        self.btn_clear = wx.Button(self.panel,label='CLEAR ALL')

        self.btn_clear.Bind(wx.EVT_BUTTON,self.clearMask)

        vbox_areas.Add(self.slct_area, flag=wx.ALL|wx.EXPAND, border=10)
        vbox_areas.Add(hbox_areas, flag=wx.ALL|wx.EXPAND, border=10)
        vbox_areas.Add(self.btn_clear, flag=wx.ALL|wx.EXPAND, border=10)
        drawbox.Add(vbox_areas, flag=wx.ALL|wx.EXPAND, border=10)

        self.toolbox.Add(drawbox, flag=wx.ALL, border=10)

        self.btn_save = wx.Button(self.panel,label='SAVE MASK')
        self.btn_save.Bind(wx.EVT_BUTTON,self.saveMask)
        self.toolbox.Add(self.btn_save, flag=wx.ALL, border=10)

        self.btn_SHWarea.Disable()
        self.btn_DELarea.Disable()
        self.btn_clear.Disable()
        self.btn_save.Disable()

    def startIMAGE(self):

        self.loadIMAGE()

    def loadIMAGE(self,event=None):
        wildcards = 'XRD image (*.edf,*.tif,*.tiff)|*.tif;*.tiff;*.edf|All files (*.*)|*.*'
        dlg = wx.FileDialog(self, message='Choose XRD image',
                           defaultDir=get_cwd(),
                           wildcard=wildcards, style=wx.FD_OPEN)

        path, read = None, False
        if dlg.ShowModal() == wx.ID_OK:
            read = True
            path = dlg.GetPath().replace('\\', '/')
        dlg.Destroy()

        if read:
            try:
#                 self.raw_img = plt.imread(path)
                self.raw_img = tifffile.imread(path)
                #self.raw_img = fabio.open(path).data
            except:
                print('Image not properly opened.')
                self.raw_img = np.zeros((1024,1024))
        else:
            print('No image selected.')
            self.raw_img = np.zeros((1024,1024))

    def ImageSizer(self):
        '''
        Image Panel
        '''
        self.imagebox = wx.BoxSizer(wx.VERTICAL)

        self.plot2Dimage()

        imagetools = wx.BoxSizer(wx.HORIZONTAL)

        self.btn_image = wx.Button(self.panel,label='IMAGE TOOLS')
        self.btn_load = wx.Button(self.panel,label='CHANGE IMAGE')

        self.btn_image.Bind(wx.EVT_BUTTON,self.onImageTools)
        self.btn_load.Bind(wx.EVT_BUTTON,self.loadIMAGE)

        imagetools.Add(self.btn_load, flag=wx.ALL, border=10)
        imagetools.Add(self.btn_image, flag=wx.ALL, border=10)


        self.imagebox.Add(self.plot2Dimg,proportion=1,flag=wx.ALL|wx.EXPAND, border=10)
        self.imagebox.Add(imagetools, flag=wx.ALL, border=10)

    def write_message(self, s, panel=0):
        """write a message to the Status Bar"""
        self.SetStatusText(s, panel)

    def onImageTools(self,event=None):

        self.toolbox = ImageToolboxFrame(self.plot2Dimg,self.raw_img)

    def plot2Dimage(self):

        self.plot2Dimg = ImagePanel(self.panel)#,size=(300, 300))
        self.plot2Dimg.messenger = self.write_message

        self.Bind(wx.EVT_PAINT, self.OnPaint)

        self.plot2Dimg.display(self.raw_img)
        self.AutoContrast()

        self.plot2Dimg.redraw()

    def AutoContrast(self):

        self.minINT = int(np.min(self.raw_img))
        self.maxINT = int(np.max(self.raw_img)/15) # /15 scales image to viewable
        if self.maxINT == self.minINT:
            self.minINT = self.minINT-50
            self.maxINT = self.minINT+100

        self.minCURRENT = self.minINT
        self.maxCURRENT = self.maxINT
        if self.maxCURRENT > self.maxINT:
            self.maxCURRENT = self.maxINT

        self.plot2Dimg.conf.auto_intensity = False
        self.plot2Dimg.conf.int_lo[0] = self.minCURRENT
        self.plot2Dimg.conf.int_hi[0] = self.maxCURRENT
#         self.plot2Dimg.conf.int_lo['int'] = self.minCURRENT
#         self.plot2Dimg.conf.int_hi['int'] = self.maxCURRENT

        ## vertical flip default
        self.plot2Dimg.conf.flip_ud = True
        self.plot2Dimg.conf.flip_lr = False

        self.plot2Dimg.redraw()

    def ShapeChoice(self,event=None):

        print('The shape you choose: %s' %  self.ch_shp.GetString(self.ch_shp.GetSelection()))

        print()
        print('Not implemented: ShapeChoice')
        self.addAREA()


    def OnPaint(self, event=None):

        print('Shape : %s' %  self.ch_shp.GetString(self.ch_shp.GetSelection()))
#         dc = wx.PaintDC(self)
#         dc.Clear()
#         dc.SetPen(wx.Pen(wx.BLACK, 4))
#         dc.DrawLine(0, 0, 50, 50)

    def clearMask(self,event=None):

        print('Clearing the mask...')
        ## provide a warning message?
        self.area_list = []
        self.slct_area.Set(self.area_list)

        self.btn_SHWarea.Disable()
        self.btn_DELarea.Disable()
        self.btn_clear.Disable()
        self.btn_save.Disable()


    def saveMask(self,event=None):

        print('This will trigger the saving of a mask.')

    def addAREA(self,event=None):
        area_name = 'area %i (%s)' % (len(self.area_list),self.ch_shp.GetString(self.ch_shp.GetSelection()))
        self.area_list.append(area_name)
        self.slct_area.Set(self.area_list)


        if len(self.area_list) > 0:
            self.btn_SHWarea.Enable()
            self.btn_DELarea.Enable()
            self.btn_clear.Enable()
            self.btn_save.Enable()

    def showAREA(self,event=None):

        if len(self.area_list) > 0:
            area_str = self.slct_area.GetString(self.slct_area.GetSelection())
            str_msg = 'Displaying: %s' % area_str
            self.write_message(str_msg,panel=0)

            ## show area on map, image

    def deleteAREA(self,event=None):

        if len(self.area_list) > 0:
            area_str = self.slct_area.GetString(self.slct_area.GetSelection())
            str_msg = 'Deleting: %s' % area_str
            self.write_message(str_msg,panel=0)

            self.area_list.remove(self.slct_area.GetString(self.slct_area.GetSelection()))
            self.slct_area.Set(self.area_list)

        if len(self.area_list) == 0:
            self.btn_SHWarea.Disable()
            self.btn_DELarea.Disable()
            self.btn_clear.Disable()
            self.btn_save.Disable()

# class diFFit_XRDmask(wx.App):
#     def __init__(self):
#         wx.App.__init__(self)
#
#     def run(self):
#         self.MainLoop()
#
#     def createApp(self):
#         frame = MaskToolsPopup(None)
#         frame.Show()
#         self.SetTopWindow(frame)
#
#     def OnInit(self):
#         self.createApp()
#         return True
#
# class DebugViewer(diFFit_XRDmask):
#     def __init__(self, **kws):
#         diFFit_XRDmask.__init__(self, **kws)
#
#     def OnInit(self):
#         #self.Init()
#         self.createApp()
#         #self.ShowInspectionTool()
#         return True
#
# if __name__ == '__main__':
#     diFFit_XRDmask().run()



#!/usr/bin/env python
"""
 json utilities for larch objects
"""
import json
import io
import numpy as np
import h5py
from datetime import datetime
from collections import namedtuple

HAS_STATE = {}
try:
    from sklearn.cross_decomposition import PLSRegression
    from sklearn.linear_model import LassoLarsCV, LassoLars, Lasso
    HAS_STATE.update({'PLSRegression': PLSRegression,
                      'LassoLarsCV':LassoLarsCV,
                      'LassoLars': LassoLars, 'Lasso': Lasso})

except ImportError:
    pass

from lmfit import Parameter, Parameters
from lmfit.model import Model, ModelResult
from lmfit.minimizer import Minimizer, MinimizerResult
from lmfit.parameter import SCIPY_FUNCTIONS

from larch import Group, isgroup, Journal, ParameterGroup

from larch.xafs import FeffitDataSet, FeffDatFile, FeffPathGroup, TransformGroup
from larch.utils.strutils import bytes2str, str2bytes, fix_varname
from larch.xafs.feffutils import FeffCalcResults

HAS_STATE['FeffCalcResults'] = FeffCalcResults
HAS_STATE['FeffDatFile'] = FeffDatFile
HAS_STATE['FeffPathGroup'] = FeffPathGroup
HAS_STATE['Journal'] = Journal

LarchGroupTypes = {'Group': Group,
                   'ParameterGroup': ParameterGroup,
                   'FeffitDataSet': FeffitDataSet,
                   'TransformGroup': TransformGroup,
                   'MinimizerResult': MinimizerResult
                   }
                   # 'FeffDatFile':  FeffDatFile,
                   # 'FeffPathGroup': FeffPathGroup,


def encode4js(obj):
    """return an object ready for json encoding.
    has special handling for many Python types
      numpy array
      complex numbers
      Larch Groups
      Larch Parameters
    """
    if obj is None:
        return None
    if isinstance(obj, np.ndarray):
        out = {'__class__': 'Array', '__shape__': obj.shape,
               '__dtype__': obj.dtype.name}
        out['value'] = obj.flatten().tolist()

        if 'complex' in obj.dtype.name:
            out['value'] = [(obj.real).tolist(), (obj.imag).tolist()]
        elif obj.dtype.name == 'object':
            out['value'] = [encode4js(i) for i in out['value']]
        return out
    elif isinstance(obj, (bool, np.bool_)):
        return bool(obj)
    elif isinstance(obj, (int, np.int64, np.int32)):
        return int(obj)
    elif isinstance(obj, (float, np.float64, np.float32)):
        return float(obj)
    elif isinstance(obj, str):
        return str(obj)
    elif isinstance(obj, bytes):
        return obj.decode('utf-8')
    elif isinstance(obj, datetime):
        return {'__class__': 'Datetime', 'isotime': obj.isoformat()}
    elif isinstance(obj,(complex, np.complex128)):
        return {'__class__': 'Complex', 'value': (obj.real, obj.imag)}
    elif isinstance(obj, io.IOBase):
        return {'__class__':  'IOBasee', 'class': obj.__class__.__name__,
                'name': obj.name, 'closed': obj.closed,
                'readable': obj.readable()}
    elif isinstance(obj, h5py.File):
        return {'__class__': 'HDF5File',
                'value': (obj.name, obj.filename, obj.mode, obj.libver),
                'keys': list(obj.keys())}
    elif isinstance(obj, h5py.Group):
        return {'__class__': 'HDF5Group', 'value': (obj.name, obj.file.filename),
                'keys': list(obj.keys())}
    elif isinstance(obj, slice):
        return {'__class__': 'Slice', 'value': (obj.start, obj.stop, obj.step)}

    elif isinstance(obj, list):
        return {'__class__': 'List', 'value': [encode4js(item) for item in obj]}
    elif isinstance(obj, tuple):
        if hasattr(obj, '_fields'):  # named tuple!
            return {'__class__': 'NamedTuple',
                    '__name__': obj.__class__.__name__,
                    '_fields': obj._fields,
                    'value': [encode4js(item) for item in obj]}
        else:
            return {'__class__': 'Tuple', 'value': [encode4js(item) for item in obj]}
    elif isinstance(obj, dict):
        out = {'__class__': 'Dict'}
        for key, val in obj.items():
            out[encode4js(key)] = encode4js(val)
        return out

    elif isinstance(obj, MinimizerResult):
        out = {'__class__': 'MinimizerResult'}
        for attr in ('aborted', 'aic', 'bic', 'call_kws', 'chisqr',
                     'covar', 'errorbars', 'ier', 'init_vals',
                     'init_values', 'last_internal_values',
                     'lmdif_message', 'message', 'method', 'ndata', 'nfev',
                     'nfree', 'nvarys', 'params', 'redchi', 'residual',
                     'success', 'var_names'):
            out[attr] = encode4js(getattr(obj, attr, None))
        return out
    elif isinstance(obj, Parameters):
        out = {'__class__': 'Parameters'}
        o_ast = obj._asteval
        out['unique_symbols'] = {key: encode4js(o_ast.symtable[key])
                                 for key in o_ast.user_defined_symbols()}
        out['params'] = [(p.name, p.__getstate__()) for p in obj.values()]
        return out
    elif isinstance(obj, Parameter):
        return {'__class__': 'Parameter', 'name': obj.name, 'state': obj.__getstate__()}
    elif isinstance(obj, Model):
        return {'__class__': 'Model', 'value': obj.dumps()}
    elif isinstance(obj, ModelResult):
        return {'__class__': 'ModelResult', 'value': obj.dumps()}
    elif hasattr(obj, '__getstate__'):
        return {'__class__': 'StatefulObject',
                '__type__': obj.__class__.__name__,
                'value': encode4js(obj.__getstate__())}

    elif isgroup(obj):
        try:
            classname = obj.__class__.__name__
        except:
            classname = 'Group'
        out = {'__class__': classname}

        if classname == 'ParameterGroup':  # save in order of parameter names
            parnames = dir(obj)
            for par in obj.__params__.keys():
                if par in parnames:
                    out[par] = encode4js(getattr(obj, par))
        else:
            for item in dir(obj):
                out[item] = encode4js(getattr(obj, item))
        return out
    elif isinstance(obj, type):
        return {'__class__': 'Type',  'value': repr(obj),
                'module': getattr(obj, '__module__', None)}
    elif callable(obj):
        return {'__class__': 'Method', '__name__': repr(obj)}
    elif hasattr(obj, 'dumps'):
        print("Encode Warning: using dumps for ", obj)
        return {'__class__': 'DumpableObject', 'value': obj.dumps()}
    else:
        print("Encode Warning: generic object dump for ", repr(obj))
        out = {'__class__': 'Object', '__repr__': repr(obj),
               '__classname__': obj.__class__.__name__}
        for attr in dir(obj):
            if attr.startswith('__') and attr.endswith('__'):
                continue
            thing = getattr(obj, attr)
            if not callable(thing):
                out[attr] = encode4js(thing)
        return out

    return obj

def decode4js(obj):
    """
    return decoded Python object from encoded object.

    """
    if not isinstance(obj, dict):
        return obj
    out = obj
    classname = obj.pop('__class__', None)
    if classname is None:
        return obj

    if classname == 'Complex':
        out = obj['value'][0] + 1j*obj['value'][1]
    elif classname in ('List', 'Tuple', 'NamedTuple'):
        out = []
        for item in obj['value']:
            out.append(decode4js(item))
        if classname == 'Tuple':
            out = tuple(out)
        elif classname == 'NamedTuple':
            out = namedtuple(obj['__name__'], obj['_fields'])(*out)
    elif classname == 'Array':
        if obj['__dtype__'].startswith('complex'):
            re = np.asarray(obj['value'][0], dtype='double')
            im = np.asarray(obj['value'][1], dtype='double')
            out = re + 1j*im
        elif obj['__dtype__'].startswith('object'):
            val = [decode4js(v) for v in obj['value']]
            out = np.array(val,  dtype=obj['__dtype__'])

        else:
            out = np.asarray(obj['value'], dtype=obj['__dtype__'])
        out.shape = obj['__shape__']
    elif classname in ('Dict', 'dict'):
        out = {}
        for key, val in obj.items():
            out[key] = decode4js(val)
    elif classname == 'Datetime':
        obj = datetime.fromisoformat(obj['isotime'])
    elif classname == 'Parameters':
        out = Parameters()
        out.clear()
        unique_symbols = {key: decode4js(obj['unique_symbols'][key]) for key
                          in obj['unique_symbols']}

        state = {'unique_symbols': unique_symbols, 'params': []}
        for name, parstate in obj['params']:
            par = Parameter(decode4js(name))
            par.__setstate__(decode4js(parstate))
            state['params'].append(par)
        out.__setstate__(state)
    elif classname in ('Parameter', 'parameter'):
        name = decode4js(obj['name'])
        state = decode4js(obj['state'])
        out = Parameter(name)
        out.__setstate__(state)

    elif classname == 'Model':
        mod = Model(lambda x: x)
        out = mod.loads(decode4js(obj['value']))

    elif classname == 'ModelResult':
        params = Parameters()
        res = ModelResult(Model(lambda x: x, None), params)
        out = res.loads(decode4js(obj['value']))

    elif classname == 'StatefulObject':
        dtype = obj.get('__type__')
        if dtype in HAS_STATE:
            out = HAS_STATE[dtype]()
            out.__setstate__(decode4js(obj.get('value')))
        else:
            print(f"Warning: cannot re-create stateful object of type '{dtype}'")

    elif classname in LarchGroupTypes:
        out = {}
        for key, val in obj.items():
            if (isinstance(val, dict) and
                val.get('__class__', None) == 'Method' and
                val.get('__name__', None) is not None):
                pass  # ignore class methods for subclassed Groups
            else:
                out[key] = decode4js(val)
        out = LarchGroupTypes[classname](**out)
    elif classname == 'Method':
        mname = obj.get('__name__', '')
        if 'ufunc' in mname:
            mname = mname.replace('<ufunc', '').replace('>', '').replace("'","").strip()
        out = SCIPY_FUNCTIONS.get(mname, None)

    else:
        print("cannot decode ", classname)
    return out



import json
import time
import numpy as np
import uuid, socket, platform
from collections import namedtuple

from gzip import GzipFile

from collections import OrderedDict

from lmfit import Parameter, Parameters
# from lmfit.model import Model, ModelResult
# from lmfit.minimizer import Minimizer, MinimizerResult

from larch import Group, isgroup, __date__, __version__, __release_version__
from ..utils import isotime, bytes2str, str2bytes, fix_varname, is_gzip
from ..utils.jsonutils import encode4js, decode4js

SessionStore = namedtuple('SessionStore', ('config', 'command_history', 'symbols'))

def get_machineid():
    "machine id / MAC address, independent of hostname"
    return hex(uuid.getnode())[2:]

def is_larch_session_file(fname):
    fopen = GzipFile if is_gzip(fname) else open
    text = 'No'
    with fopen(fname, 'rb') as fh:
        text = fh.read(64).decode('utf-8')
    return text.startswith('##LARIX:')

def save_groups(fname, grouplist):
    """save a list of groups (and other supported datatypes) to file

    This is a simplified and minimal version of save_session()

    Use 'read_groups()' to read data saved from this function
    """
    buff = ["##LARCH GROUPLIST"]
    for dat in grouplist:
        buff.append(json.dumps(encode4js(dat)))

    buff.append("")

    fh = GzipFile(fname, "w")
    fh.write(str2bytes("\n".join(buff)))
    fh.close()

def read_groups(fname):
    """read a list of groups (and other supported datatypes)
    from a file saved with 'save_groups()'

    Returns a list of objects
    """
    fopen = GzipFile if is_gzip(fname) else open
    with fopen(fname, 'rb') as fh:
        text = fh.read().decode('utf-8')

    lines = text.split('\n')
    line0 = lines.pop(0)
    if not line0.startswith('##LARCH GROUPLIST'):
        raise ValueError(f"Invalid Larch group file: '{fname:s}'")

    out = []
    for line in lines:
        if len(line) > 1:
            out.append(decode4js(json.loads(line)))
    return out


def save_session(fname=None, _larch=None):
    """save all groups and data into a Larch Save File (.larix)
    A portable json file, that can be loaded with

    load_session(fname)

    Parameters
    ----------
    fname   name of output save file.

    See Also:  restore_session()
    """
    if fname is None:
        fname = time.strftime('%Y%b%d_%H%M')
    if not fname.endswith('.larix'):
        fname = fname + '.larix'

    if _larch is None:
        raise ValueError('_larch not defined')
    symtab = _larch.symtable

    buff = ["##LARIX: 1.0      Larch Session File",
            "##Date Saved: %s"   % time.strftime('%Y-%m-%d %H:%M:%S'),
            "##<CONFIG>",
            "##Machine Platform: %s" % platform.system(),
            "##Machine Name: %s" % socket.gethostname(),
            "##Machine MACID: %s" % get_machineid(),
            "##Machine Version: %s"   % platform.version(),
            "##Machine Processor: %s" % platform.machine(),
            "##Machine Architecture: %s" % ':'.join(platform.architecture()),
            "##Python Version: %s" % platform.python_version(),
            "##Python Compiler: %s" % platform.python_compiler(),
            "##Python Implementation: %s" % platform.python_implementation(),
            "##Larch Release Version: %s" % __release_version__,
            "##Larch Release Date: %s" % __date__,
            ]

    core_groups = symtab._sys.core_groups
    buff.append('##Larch Core Groups: %s' % (json.dumps(core_groups)))

    config = symtab._sys.config
    for attr in dir(config):
        buff.append('##Larch %s: %s' % (attr, json.dumps(getattr(config, attr, None))))
    buff.append("##</CONFIG>")

    try:
        histbuff = _larch.input.history.get(session_only=True)
    except:
        histbuff = None

    if histbuff is not None:
        buff.append("##<Session Commands>")
        buff.extend(["%s" % l for l in histbuff])
        buff.append("##</Session Commands>")

    syms = []
    for attr in symtab.__dir__(): # insert order, not alphabetical order
        if attr in core_groups:
            continue
        syms.append(attr)
    buff.append("##<Symbols: count=%d>"  % len(syms))

    for attr in dir(symtab):
        if attr in core_groups:
            continue
        buff.append('<:%s:>' % attr)
        buff.append('%s' % json.dumps(encode4js(getattr(symtab, attr))))

    buff.append("##</Symbols>")
    buff.append("")

    fh = GzipFile(fname, "w")
    fh.write(str2bytes("\n".join(buff)))
    fh.close()

def read_session(fname):
    """read Larch Save File, returning data

    Arguments
    ---------
    fname    name of save file

    Returns
    -------
    symbols (dict), configuration (dict), command history (list)

    """
    fopen = GzipFile if is_gzip(fname) else open
    with fopen(fname, 'rb') as fh:
        text = fh.read().decode('utf-8')

    lines = text.split('\n')
    line0 = lines.pop(0)
    if not line0.startswith('##LARIX:'):
        raise ValueError(f"Invalid Larch session file: '{fname:s}'")

    version = line0.split()[1]

    symbols = {}
    config = {'Larix Version': version}
    cmd_history = []
    nsyms = nsym_expected = 0
    section = symname = '_unknown_'

    for line in lines:
        if line.startswith("##<"):
            section = line.replace('##<','').replace('>', '').strip().lower()
            if ':' in section:
                section, options = section.split(':', 1)
            if section.startswith('/'):
                section = '_unknown_'
        elif section == 'session commands':
            cmd_history.append(line)

        elif section == 'symbols':
            if line.startswith('<:') and line.endswith(':>'):
                symname = line.replace('<:', '').replace(':>', '')
            else:
                symbols[symname] = decode4js(json.loads(line))
        else:
            if line.startswith('##') and ':' in line:
                line = line[2:]
                key, val = line.split(':', 1)
                key = key.strip()
                val = val.strip()
                if '[' in val or '{' in val:
                    try:
                        val = decode4js(json.loads(val))
                    except:
                        pass
                config[key] = val

    return SessionStore(config, cmd_history, symbols)


def load_session(fname, overwrite=True, merge_dicts=True, _larch=None):
    """load all data from a Larch Save File into current larch session

    Arguments
    ---------
    fname  (str)     name of save file
    overwrite (bool) whether to overwrite existing symbols [True]

    Returns
    -------
    None, puts data into current session

    """
    if _larch is None:
        raise ValueError('load session needs a larch session')

    session = read_session(fname)

    symtab = _larch.symtable
    if not hasattr(symtab._sys, 'restored_sessions'):
        symtab._sys.restored_sessions = {}
    this = symtab._sys.restored_sessions[fname] = {}
    this['date'] = isotime()
    this['config'] = session.config
    this['command_history'] = session.command_history


    for sym, val in session.symbols.items():
        cur = getattr(symtab, sym, None)
        if isinstance(cur, dict) and merge_dicts:
            cur.update(val)
            setattr(symtab, sym, cur)
        elif overwrite or cur is None:
            setattr(symtab, sym, val)



#!/usr/bin/env python
"""
Linear Combination panel
"""
import os
import time
import wx
import wx.grid as wxgrid
import numpy as np
import pickle
import base64
from copy import deepcopy
from functools import partial

from lmfit.printfuncs import gformat
from larch import Group
from larch.math import index_of
from larch.wxlib import (BitmapButton, TextCtrl, FloatCtrl, get_icon,
                         SimpleText, pack, Button, HLine, Choice, Check,
                         NumericCombo, CEN, LEFT, Font, FileSave, FileOpen,
                         DataTableGrid, Popup)
from larch.io import save_groups, read_groups, read_csv
from larch.utils.strutils import fix_varname
from larch.utils import get_cwd

from .taskpanel import TaskPanel
from .config import Linear_ArrayChoices, Regress_Choices

CSV_WILDCARDS = "CSV Files(*.csv,*.dat)|*.csv*;*.dat|All files (*.*)|*.*"
MODEL_WILDCARDS = "Regression Model Files(*.regmod,*.dat)|*.regmod*;*.dat|All files (*.*)|*.*"

Plot_Choices = ['Mean Spectrum + Active Energies',
                'Spectra Stack',
                'Predicted External Varliable']

MAX_ROWS = 1000

def make_steps(max=1, decades=8):
    steps = [1.0]
    for i in range(6):
        steps.extend([(j*10**(-(1+i))) for j in (5, 2, 1)])
    return steps

class RegressionPanel(TaskPanel):
    """Regression Panel"""
    def __init__(self, parent, controller, **kws):
        TaskPanel.__init__(self, parent, controller, panel='regression', **kws)
        self.result = None
        self.save_csvfile   = 'RegressionData.csv'
        self.save_modelfile = 'Model.regmod'

    def process(self, dgroup, **kws):
        """ handle processing"""
        if self.skip_process:
            return
        self.skip_process = True
        form = self.read_form()

    def build_display(self):
        panel = self.panel
        wids = self.wids
        self.skip_process = True

        wids['fitspace'] = Choice(panel, choices=list(Linear_ArrayChoices.keys()),
                                  action=self.onFitSpace, size=(175, -1))
        wids['fitspace'].SetSelection(0)
        # wids['plotchoice'] = Choice(panel, choices=Plot_Choices,
        #                           size=(250, -1), action=self.onPlot)

        wids['method'] = Choice(panel, choices=Regress_Choices, size=(250, -1),
                                action=self.onRegressMethod)
        wids['method'].SetSelection(1)
        add_text = self.add_text

        opts = dict(digits=2, increment=1.0)
        defaults = self.get_defaultconfig()

        self.make_fit_xspace_widgets(elo=defaults['elo_rel'], ehi=defaults['ehi_rel'])

        wids['alpha'] =  NumericCombo(panel, make_steps(), fmt='%.6g',
                                      default_val=0.01, width=100)

        wids['auto_scale_pls'] = Check(panel, default=True, label='auto scale?')
        wids['auto_alpha'] = Check(panel, default=False, label='auto alpha?')

        wids['fit_intercept'] = Check(panel, default=True, label='fit intercept?')

        wids['save_csv'] = Button(panel, 'Save CSV File', size=(150, -1),
                                    action=self.onSaveCSV)
        wids['load_csv'] = Button(panel, 'Load CSV File', size=(150, -1),
                                    action=self.onLoadCSV)

        wids['save_model'] = Button(panel, 'Save Model', size=(150, -1),
                                    action=self.onSaveModel)
        wids['save_model'].Disable()

        wids['load_model'] = Button(panel, 'Load Model', size=(150, -1),
                                    action=self.onLoadModel)


        wids['train_model'] = Button(panel, 'Train Model From These Data',
                                     size=(275, -1),  action=self.onTrainModel)

        wids['fit_group'] = Button(panel, 'Predict Variable for Selected Groups',
                                   size=(275, -1), action=self.onPredictGroups)
        wids['fit_group'].Disable()


        w_cvfolds = self.add_floatspin('cv_folds', digits=0, with_pin=False,
                                       value=0, increment=1, min_val=-1)

        w_cvreps  = self.add_floatspin('cv_repeats', digits=0, with_pin=False,
                                       value=0, increment=1, min_val=-1)

        w_ncomps  = self.add_floatspin('ncomps', digits=0, with_pin=False,
                                       value=3, increment=1, min_val=1)

        wids['varname'] = wx.TextCtrl(panel, -1, 'valence', size=(150, -1))
        wids['stat1'] =  SimpleText(panel, ' - - - ')
        wids['stat2'] =  SimpleText(panel, ' - - - ')


        collabels = [' File /Group Name ', 'External Value',
                     'Predicted Value', 'Training?']
        colsizes = [250, 100, 100, 100]
        coltypes = ['str', 'float:12,4', 'float:12,4', 'str']
        coldefs  = ['', 0.0, 0.0, '']

        wids['table'] = DataTableGrid(panel, nrows=MAX_ROWS,
                                      collabels=collabels,
                                      datatypes=coltypes,
                                      defaults=coldefs,
                                      colsizes=colsizes)

        wids['table'].SetMinSize((650, 225))

        wids['use_selected'] = Button(panel, 'Use Selected Groups',
                                      size=(150, -1),  action=self.onFillTable)

        panel.Add(SimpleText(panel, 'Feature Regression, Model Selection',
                             size=(350, -1), **self.titleopts), style=LEFT, dcol=4)

        add_text('Array to Use: ', newrow=True)
        panel.Add(wids['fitspace'], dcol=4)

        panel.Add(wids['fitspace_label'], newrow=True)
        panel.Add(self.elo_wids)
        add_text(' : ', newrow=False)
        panel.Add(self.ehi_wids, dcol=3)
        add_text('Regression Method:')
        panel.Add(wids['method'], dcol=4)
        add_text('PLS # components: ')
        panel.Add(w_ncomps)
        panel.Add(wids['auto_scale_pls'], dcol=2)
        add_text('Lasso Alpha: ')
        panel.Add(wids['alpha'])
        panel.Add(wids['auto_alpha'], dcol=2)
        panel.Add(wids['fit_intercept'])

        add_text('Cross Validation: ')
        add_text(' # folds, # repeats: ', newrow=False)
        panel.Add(w_cvfolds, dcol=2)
        panel.Add(w_cvreps)

        panel.Add(HLine(panel, size=(600, 2)), dcol=6, newrow=True)

        add_text('Build Model: ', newrow=True)
        panel.Add(wids['use_selected'],   dcol=2)
        add_text('Attribute Name: ', newrow=False)
        panel.Add(wids['varname'], dcol=4)

        add_text('Read/Save Data: ', newrow=True)
        panel.Add(wids['load_csv'], dcol=3)
        panel.Add(wids['save_csv'], dcol=2)

        panel.Add(wids['table'], newrow=True, dcol=5) # , drow=3)

        panel.Add(HLine(panel, size=(550, 2)), dcol=5, newrow=True)
        panel.Add((5, 5), newrow=True)
        add_text('Train Model : ')
        panel.Add(wids['train_model'], dcol=3)
        panel.Add(wids['load_model'])

        add_text('Use This Model : ')
        panel.Add(wids['fit_group'], dcol=3)
        panel.Add(wids['save_model'])
        add_text('Fit Statistics : ')
        panel.Add(wids['stat1'], dcol=4)
        panel.Add((5, 5), newrow=True)
        panel.Add(wids['stat2'], dcol=4)
        panel.pack()

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add((10, 10), 0, LEFT, 3)
        sizer.Add(panel, 1, LEFT, 3)
        pack(self, sizer)
        self.onRegressMethod()
        self.skip_process = False

    def onRegressMethod(self, evt=None):
        meth = self.wids['method'].GetStringSelection()
        use_lasso = meth.lower().startswith('lasso')
        self.wids['alpha'].Enable(use_lasso)
        self.wids['auto_alpha'].Enable(use_lasso)
        self.wids['fit_intercept'].Enable(use_lasso)
        self.wids['auto_scale_pls'].Enable(not use_lasso)
        self.wids['ncomps'].Enable(not use_lasso)

    def onFitSpace(self, evt=None):
        fitspace = self.wids['fitspace'].GetStringSelection()
        self.update_config(dict(fitspace=fitspace))
        arrname = Linear_ArrayChoices.get(fitspace, 'norm')
        self.update_fit_xspace(arrname)
        self.plot()


    def fill_form(self, dgroup=None, opts=None):
        conf = deepcopy(self.get_config(dgroup=dgroup, with_erange=True))
        if opts is None:
            opts = {}
        conf.update(opts)
        self.dgroup = dgroup
        self.skip_process = True
        wids = self.wids

        for attr in ('fitspace','method'):
            if attr in conf:
                wids[attr].SetStringSelection(conf[attr])

        for attr in ('elo', 'ehi', 'alpha', 'variable', 'cv_folds', 'cv_repeats'):
            val = conf.get(attr, None)
            if val is not None:
                if attr == 'alpha':
                    if val < 0:
                        val = 0.001
                        conf['auto_alpha'] = True
                    val = '%.6g' % val
                wids[attr].SetValue(val)

        use_lasso = conf['method'].lower().startswith('lasso')

        for attr in ('auto_alpha', 'fit_intercept','auto_scale_pls'):
            val = conf.get(attr, True)
            if attr == 'auto_scale_pls':
                val = val and not use_lasso
            else:
                val = val and use_lasso
            wids[attr].SetValue(val)
        self.onRegressMethod()

        self.skip_process = False

    def read_form(self):
        dgroup = self.controller.get_group()
        form = {'groupname': getattr(dgroup, 'groupname', 'No Group')}

        for k in ('fitspace', 'method'):
            form[k] = self.wids[k].GetStringSelection()

        for k in ('elo', 'ehi', 'alpha', 'cv_folds',
                  'cv_repeats', 'ncomps', 'varname'):
            form[k] = self.wids[k].GetValue()

        form['alpha'] = float(form['alpha'])
        if form['alpha'] < 0:
            form['alpha'] = 1.e-3

        for k in ('auto_scale_pls', 'auto_alpha',  'fit_intercept'):
            form[k] = self.wids[k].IsChecked()

        mname = form['method'].lower()
        form['use_lars'] = 'lars' in mname
        form['funcname'] = 'pls'
        if mname.startswith('lasso'):
            form['funcname'] = 'lasso'
            if form['auto_alpha']:
                form['alpha'] = None

        return form


    def onFillTable(self, event=None):
        selected_groups = self.controller.filelist.GetCheckedStrings()
        varname = fix_varname(self.wids['varname'].GetValue())
        predname = varname + '_predicted'
        grid_data = []
        for fname in self.controller.filelist.GetCheckedStrings():
            gname = self.controller.file_groups[fname]
            grp = self.controller.get_group(gname)
            grid_data.append([fname, getattr(grp, varname, 0.0),
                              getattr(grp, predname, 0.0), 'Yes'])

        self.wids['table'].table.data = grid_data
        self.wids['table'].table.View.Refresh()

    def onTrainModel(self, event=None):
        form = self.read_form()
        varname = form['varname']
        predname = varname + '_predicted'

        grid_data = self.wids['table'].table.data
        groups = []
        for fname, yval, pval, istrain in grid_data:
            gname = self.controller.file_groups[fname]
            grp = self.controller.get_group(gname)
            setattr(grp, varname, yval)
            setattr(grp, predname, pval)
            groups.append(gname)

        cmds = ['# train linear regression model',
               'training_groups = [%s]' % ', '.join(groups)]

        copts = ["varname='%s'" % varname, "xmin=%.4f" % form['elo'],
                 "xmax=%.4f" % form['ehi']]

        arrname = Linear_ArrayChoices.get(form['fitspace'], 'norm')
        copts.append("arrayname='%s'" % arrname)

        if form['method'].lower().startswith('lasso'):
            if form['auto_alpha']:
                copts.append('alpha=None')
            else:
                copts.append('alpha=%.6g' % form['alpha'])
            copts.append('use_lars=%s' % repr('lars' in form['method'].lower()))
            copts.append('fit_intercept=%s' % repr(form['fit_intercept']))
        else:
            copts.append('ncomps=%d' % form['ncomps'])
            copts.append('scale=%s' % repr(form['auto_scale_pls']))

        callargs = ', '.join(copts)

        cmds.append("reg_model = %s_train(training_groups, %s)" %
                    (form['funcname'], callargs))

        self.larch_eval('\n'.join(cmds))
        reg_model = self.larch_get('reg_model')
        reg_model.form = form
        self.use_regmodel(reg_model)

    def use_regmodel(self, reg_model):
        if reg_model is None:
            return
        opts = self.read_form()

        if hasattr(reg_model, 'form'):
            opts.update(reg_model.form)

        self.write_message('Regression Model trained: %s' % opts['method'])
        rmse_cv = reg_model.rmse_cv
        if rmse_cv is not None:
            rmse_cv = "%.4f" % rmse_cv
        stat = "RMSE_CV = %s, RMSE = %.4f" % (rmse_cv, reg_model.rmse)
        self.wids['stat1'].SetLabel(stat)
        if opts['funcname'].startswith('lasso'):
            stat = "Alpha = %.4f, %d active components"
            self.wids['stat2'].SetLabel(stat % (reg_model.alpha,
                                                len(reg_model.active)))

            if opts['auto_alpha']:
                self.wids['alpha'].add_choice(reg_model.alpha)

        else:
            self.wids['stat2'].SetLabel('- - - ')
        training_groups = reg_model.groupnames
        ntrain = len(training_groups)
        grid_data = self.wids['table'].table.data
        grid_new = []
        for i in range(ntrain): # min(ntrain, len(grid_data))):
            fname = training_groups[i]
            istrain = 'Yes' if fname in training_groups else 'No'
            grid_new.append( [fname, reg_model.ydat[i], reg_model.ypred[i], istrain])
        self.wids['table'].table.data = grid_new
        self.wids['table'].table.View.Refresh()

        if reg_model.cv_folds not in (0, None):
            self.wids['cv_folds'].SetValue(reg_model.cv_folds)
        if reg_model.cv_repeats not in (0, None):
            self.wids['cv_repeats'].SetValue(reg_model.cv_repeats)

        self.wids['save_model'].Enable()
        self.wids['fit_group'].Enable()

        wx.CallAfter(self.onPlotModel, model=reg_model)

    def onPanelExposed(self, **kws):
        # called when notebook is selected
        try:
            fname = self.controller.filelist.GetStringSelection()
            gname = self.controller.file_groups[fname]
            dgroup = self.controller.get_group(gname)
            if not hasattr(dgroup, 'norm'):
                self.xasmain.process_normalization(dgroup)
            self.fill_form(dgroup)
        except:
            pass # print(" Cannot Fill prepeak panel from group ")

        reg_model = getattr(self.larch.symtable, 'reg_model', None)
        if reg_model is not None:
            self.use_regmodel(reg_model)


    def onPredictGroups(self, event=None):
        opts = self.read_form()
        varname = opts['varname'] + '_predicted'

        reg_model = self.larch_get('reg_model')
        training_groups = reg_model.groupnames

        grid_data = self.wids['table'].table.data

        gent = {}
        if len(grid_data[0][0].strip()) == 0:
            grid_data = []
        else:
            for i, row in enumerate(grid_data):
                gent[row[0]] = i

        for fname in self.controller.filelist.GetCheckedStrings():
            gname = self.controller.file_groups[fname]
            grp   = self.controller.get_group(gname)
            extval = getattr(grp, opts['varname'], 0)
            cmd = "%s.%s = %s_predict(%s, reg_model)" % (gname, varname,
                                                         opts['funcname'], gname)
            self.larch_eval(cmd)
            val = self.larch_get('%s.%s' % (gname, varname))
            if fname in gent:
                grid_data[gent[fname]][2] = val
            else:
                istrain = 'Yes' if fname in training_groups else 'No'
                grid_data.append([fname, extval, val, istrain])
            self.wids['table'].table.data = grid_data
            self.wids['table'].table.View.Refresh()

    def onSaveModel(self, event=None):
        try:
            reg_model = self.larch_get('reg_model')
        except:
            reg_model = None
        if reg_model is None:
            self.write_message('Cannot Save Regression Model')
            return

        dlg = wx.FileDialog(self, message="Save Regression Model",
                            defaultDir=get_cwd(),
                            defaultFile=self.save_modelfile,
                            wildcard=MODEL_WILDCARDS,
                            style=wx.FD_SAVE)
        fname = None
        if dlg.ShowModal() == wx.ID_OK:
            fname = dlg.GetPath()
        dlg.Destroy()
        if fname is None:
            return
        save_groups(fname, ['#regression model 1.0', reg_model])
        self.write_message('Wrote Regression Model to %s ' % fname)

    def onLoadModel(self, event=None):
        dlg = wx.FileDialog(self, message="Load Regression Model",
                            defaultDir=get_cwd(),
                            wildcard=MODEL_WILDCARDS, style=wx.FD_OPEN)

        fname = None
        if dlg.ShowModal() == wx.ID_OK:
            fname = dlg.GetPath()
        dlg.Destroy()
        if fname is None:
            return
        dat = read_groups(fname)
        if len(dat) != 2 or not dat[0].startswith('#regression model'):
            Popup(self, f" '{rfile}' is not a valid Regression model file",
                  "Invalid file")

        reg_model = dat[1]
        self.controller.symtable.reg_model = reg_model

        self.write_message('Read Regression Model from %s ' % fname)
        self.wids['fit_group'].Enable()

        self.use_regmodel(reg_model)

    def onLoadCSV(self, event=None):
        dlg = wx.FileDialog(self, message="Load CSV Data File",
                            defaultDir=get_cwd(),
                            wildcard=CSV_WILDCARDS, style=wx.FD_OPEN)

        fname = None
        if dlg.ShowModal() == wx.ID_OK:
            fname = dlg.GetPath()
        dlg.Destroy()
        if fname is None:
            return

        self.save_csvfile = os.path.split(fname)[1]
        varname = fix_varname(self.wids['varname'].GetValue())
        csvgroup = read_csv(fname)
        script = []
        grid_data = []
        for sname, yval in zip(csvgroup.col_01, csvgroup.col_02):
            if sname.startswith('#'):
                continue
            if sname in self.controller.file_groups:
                gname = self.controller.file_groups[sname]
                script.append('%s.%s = %f' % (gname, varname, yval))
                grid_data.append([sname, yval, 0])

        self.larch_eval('\n'.join(script))
        self.wids['table'].table.data = grid_data
        self.wids['table'].table.View.Refresh()
        self.write_message('Read CSV File %s ' % fname)

    def onSaveCSV(self, event=None):
        wildcard = 'CSV file (*.csv)|*.csv|All files (*.*)|*.*'
        fname = FileSave(self, message='Save CSV Data File',
                         wildcard=wildcard,
                         default_file=self.save_csvfile)
        if fname is None:
            return
        self.save_csvfile = os.path.split(fname)[1]
        buff = []
        for  row in self.wids['table'].table.data:
            buff.append("%s, %s, %s" % (row[0], gformat(row[1]), gformat(row[2])))
        buff.append('')
        with open(fname, 'w') as fh:
            fh.write('\n'.join(buff))
        self.write_message('Wrote CSV File %s ' % fname)

    def onPlotModel(self, event=None, model=None):
        opts = self.read_form()
        if model is None:
            return
        opts.update(model.form)

        ppanel = self.controller.get_display(win=1).panel
        viewlims = ppanel.get_viewlimits()
        plotcmd = ppanel.plot

        d_ave = model.spectra.mean(axis=0)
        d_std = model.spectra.std(axis=0)
        ymin, ymax = (d_ave-d_std).min(), (d_ave+d_std).max()

        if opts['funcname'].startswith('lasso'):
            active = [int(i) for i in model.active]
            active_coefs = (model.coefs[active])
            active_coefs = active_coefs/max(abs(active_coefs))
            ymin = min(active_coefs.min(), ymin)
            ymax = max(active_coefs.max(), ymax)

        else:
            ymin = min(model.coefs.min(), ymin)
            ymax = max(model.coefs.max(), ymax)

        ymin = ymin - 0.02*(ymax-ymin)
        ymax = ymax + 0.02*(ymax-ymin)


        title = '%s Regression results' % (opts['method'])

        ppanel.plot(model.x, d_ave, win=1, title=title,
                    label='mean spectra', xlabel='Energy (eV)',
                    ylabel=opts['fitspace'], show_legend=True,
                    ymin=ymin, ymax=ymax)
        ppanel.axes.fill_between(model.x, d_ave-d_std, d_ave+d_std,
                                 color='#1f77b433')
        if opts['funcname'].startswith('lasso'):
            ppanel.axes.bar(model.x[active], active_coefs,
                            1.0, color='#9f9f9f88',
                            label='coefficients')
        else:
            _, ncomps = model.coefs.shape
            for i in range(ncomps):
                ppanel.oplot(model.x, model.coefs[:, i], label='coef %d' % (i+1))

        ppanel.canvas.draw()

        ngoups = len(model.groupnames)
        indices = np.arange(len(model.groupnames))
        diff = model.ydat - model.ypred
        sx = np.argsort(model.ydat)

        ppanel = self.controller.get_display(win=2).panel

        ppanel.plot(model.ydat[sx], indices, xlabel='valence',
                    label='experimental', linewidth=0, marker='o',
                    markersize=8, win=2, new=True, title=title)

        ppanel.oplot(model.ypred[sx], indices, label='predicted',
                    labelxsxfontsize=7, markersize=6, marker='o',
                    linewidth=0, show_legend=True, new=False)

        ppanel.axes.barh(indices, diff[sx], 0.5, color='#9f9f9f88')
        ppanel.axes.set_yticks(indices)
        ppanel.axes.set_yticklabels([model.groupnames[o] for o in sx])
        ppanel.conf.auto_margins = False
        ppanel.conf.set_margins(left=0.35, right=0.05, bottom=0.15, top=0.1)
        ppanel.canvas.draw()


    def onCopyParam(self, name=None, evt=None):
        conf = self.get_config()



import time
import os
import sys
import platform
from functools import partial
from copy import deepcopy

import numpy as np
np.seterr(all='ignore')

import wx
import wx.grid as wxgrid
import wx.lib.scrolledpanel as scrolled

from larch import Group
from larch.wxlib import (BitmapButton, SetTip, GridPanel, FloatCtrl,
                         FloatSpin, FloatSpinWithPin, get_icon, SimpleText,
                         pack, Button, HLine, Choice, Check, MenuItem,
                         GUIColors, CEN, LEFT, FRAMESTYLE, Font, FileSave,
                         FileOpen, FONTSIZE, FONTSIZE_FW, DataTableGrid)

from larch.xafs import etok, ktoe
from larch.utils import group2dict
from larch.utils.strutils import break_longstring
from .config import PANELS

LEFT = wx.ALIGN_LEFT
CEN |=  wx.ALL

def autoset_fs_increment(wid, value):
    """set increment for floatspin to be
    1, 2, or 5 x 10^(integer) and ~0.02 X current value
    """
    if abs(value) < 1.e-20:
        return
    ndig = int(1-round(np.log10(abs(value*0.5))))
    wid.SetDigits(ndig+1)
    c, inc = 0, 10.0**(-ndig)
    while (inc/abs(value) > 0.02):
        scale = 0.5 if (c % 2 == 0) else 0.4
        inc *= scale
        c += 1
    wid.SetIncrement(inc)


class GroupJournalFrame(wx.Frame):
    """ edit parameters"""
    def __init__(self, parent, dgroup=None, xasmain=None, **kws):
        self.xasmain = xasmain
        self.dgroup = dgroup
        self.n_entries = 0
        wx.Frame.__init__(self, None, -1,  'Group Journal',
                          style=FRAMESTYLE, size=(950, 700))

        panel = GridPanel(self, ncols=3, nrows=10, pad=2, itemstyle=LEFT)

        self.label = SimpleText(panel, 'Group Journal', size=(750, 30))

        export_btn = Button(panel, 'Export to Tab-Separated File', size=(225, -1),
                            action=self.export)

        add_btn = Button(panel, 'Add Entry', size=(200, -1), action=self.add_entry)
        self.label_wid = wx.TextCtrl(panel, -1, value='user comment', size=(200, -1))
        self.value_wid = wx.TextCtrl(panel, -1, value='',             size=(600, -1))

        panel.Add(self.label, dcol=3, style=LEFT)

        panel.Add(SimpleText(panel, 'Add a Journal Entry:'), dcol=1, style=LEFT, newrow=True)
        panel.Add(add_btn, dcol=1)
        panel.Add(export_btn, dcol=1, newrow=False)

        panel.Add(SimpleText(panel, 'Label:'), style=LEFT, newrow=True)
        panel.Add(self.label_wid, dcol=1, style=LEFT)

        panel.Add(SimpleText(panel, 'Value:'), style=LEFT, newrow=True)
        panel.Add(self.value_wid, dcol=2, style=LEFT)
        panel.pack()


        collabels = [' Label ', ' Value ', 'Date/Time']

        colsizes = [150, 550, 150]
        coltypes = ['string', 'string', 'string']
        coldefs  = [' ', ' ', ' ']

        self.datagrid = DataTableGrid(self, nrows=80,
                                      collabels=collabels,
                                      datatypes=coltypes,
                                      defaults=coldefs,
                                      colsizes=colsizes,
                                      rowlabelsize=40)

        self.datagrid.SetMinSize((925, 650))
        self.datagrid.EnableEditing(False)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(panel, 0, LEFT, 3)
        sizer.Add(self.datagrid, 1, LEFT|wx.ALL, 3)
        pack(self, sizer)

        self.xasmain.timers['journal_updater'] = wx.Timer(self.xasmain)
        self.xasmain.Bind(wx.EVT_TIMER, self.onRefresh,
                          self.xasmain.timers['journal_updater'])
        self.Bind(wx.EVT_CLOSE,  self.onClose)
        self.SetSize((950, 725))
        self.Show()
        self.Raise()
        self.xasmain.timers['journal_updater'].Start(1000)

        if dgroup is not None:
            wx.CallAfter(self.set_group, dgroup=dgroup)

    def add_entry(self, evt=None):
        if self.dgroup is not None:
            label = self.label_wid.GetValue()
            value = self.value_wid.GetValue()
            if len(label)>0 and len(value)>1:
                self.dgroup.journal.add(label, value)


    def onClose(self, event=None):
        self.xasmain.timers['journal_updater'].Stop()
        self.Destroy()

    def onRefresh(self, event=None):
        if self.dgroup is None:
            return
        if self.n_entries == len(self.dgroup.journal.data):
            return
        self.set_group(self.dgroup)


    def export(self, event=None):
        wildcard = 'CSV file (*.csv)|*.csv|All files (*.*)|*.*'
        fname = FileSave(self, message='Save Tab-Separated-Value Data File',
                         wildcard=wildcard,
                         default_file= f"{self.dgroup.filename}_journal.csv")
        if fname is None:
            return

        buff = ['Label\tValue\tDateTime']
        for entry in self.dgroup.journal:
            k, v, dt = entry.key, entry.value, entry.datetime.isoformat()
            k = k.replace('\t', '_')
            if not isinstance(v, str): v = repr(v)
            v = v.replace('\t', '   ')
            buff.append(f"{k}\t{v}\t{dt}")

        buff.append('')
        with open(fname, 'w') as fh:
            fh.write('\n'.join(buff))

        msg = f"Exported journal for {self.dgroup.filename} to '{fname}'"
        writer = getattr(self.xasmain, 'write_message', sys.stdout)
        writer(msg)


    def set_group(self, dgroup=None):
        if dgroup is None:
            dgroup = self.dgroup
        if dgroup is None:
            return
        self.dgroup = dgroup
        self.SetTitle(f'Group Journal for {dgroup.filename:s}')

        label = f'Journal for {dgroup.filename}'
        desc = dgroup.journal.get('source_desc')
        if desc is not None:
            label = f'Journal for {desc.value}'
        self.label.SetLabel(label)

        grid_data = []
        rowsize = []
        self.n_entries = len(dgroup.journal.data)

        for entry in dgroup.journal:
            val = entry.value
            if not isinstance(val, str):
                val = repr(val)
            xval = break_longstring(val)
            val = '\n'.join(xval)
            rowsize.append(len(xval))

            xtime = entry.datetime.strftime("%Y/%m/%d %H:%M:%S")
            grid_data.append([entry.key, val, xtime])

        self.datagrid.table.Clear()
        nrows = self.datagrid.table.GetRowsCount()
        if len(grid_data) > nrows:
            self.datagrid.table.AppendRows(len(grid_data)+8 - nrows)

        self.datagrid.table.data = grid_data
        for i, rsize in enumerate(rowsize):
            self.datagrid.SetRowSize(i, rsize*20)

        self.datagrid.table.View.Refresh()


class TaskPanel(wx.Panel):
    """generic panel for main tasks.   meant to be subclassed
    """
    def __init__(self, parent, controller, xasmain=None, panel=None, **kws):
        wx.Panel.__init__(self, parent, -1, size=(550, 625), **kws)
        self.parent = parent
        self.xasmain = xasmain or parent
        self.controller = controller
        self.larch = controller.larch
        self.title = 'Generic Panel'
        self.configname = None
        if panel in PANELS:
            self.configname = panel
            self.title = PANELS[panel]

        self.wids = {}
        self.subframes = {}
        self.command_hist = []
        self.SetFont(Font(FONTSIZE))
        self.titleopts = dict(font=Font(FONTSIZE+2),
                              colour='#AA0000', style=LEFT)

        self.font_fixedwidth = wx.Font(FONTSIZE_FW, wx.MODERN, wx.NORMAL, wx.NORMAL)

        self.panel = GridPanel(self, ncols=7, nrows=10, pad=2, itemstyle=LEFT)
        self.panel.sizer.SetVGap(5)
        self.panel.sizer.SetHGap(5)
        self.skip_process = True
        self.skip_plotting = False
        self.build_display()
        self.skip_process = False
        self.fit_xspace = 'e'
        self.fit_last_erange = None

    def make_fit_xspace_widgets(self, elo=-1, ehi=1):
        self.wids['fitspace_label'] = SimpleText(self.panel, 'Fit Range (eV):')
        opts = dict(digits=2, increment=1.0, relative_e0=True)
        self.elo_wids = self.add_floatspin('elo', value=elo, **opts)
        self.ehi_wids = self.add_floatspin('ehi', value=ehi, **opts)

    def update_fit_xspace(self, arrayname):
        fit_xspace = 'e'
        if arrayname.startswith('chi'):
            fit_xspace = 'r' if 'r' in arrayname else 'k'

        if fit_xspace == self.fit_xspace:
            return

        if self.fit_xspace == 'e' and fit_xspace == 'k': # e to k
            dgroup = self.controller.get_group()
            e0 = getattr(dgroup, 'e0', None)
            k  = getattr(dgroup, 'k', None)
            if e0 is None or k is None:
                return
            elo = self.wids['elo'].GetValue()
            ehi = self.wids['ehi'].GetValue()
            self.fit_last_erange = (elo, ehi)
            self.wids['elo'].SetValue(etok(elo-e0))
            self.wids['ehi'].SetValue(etok(ehi+e0))
            self.fit_xspace = 'k'
            self.wids['fitspace_label'].SetLabel('Fit Range (1/\u212B):')
        elif self.fit_xspace == 'k' and fit_xspace == 'e': # k to e
            if self.fit_last_erange is not None:
                elo, ehi = self.fit_last_erange
            else:
                dgroup = self.controller.get_group()
                e0 = getattr(dgroup, 'e0', None)
                k  = getattr(dgroup, 'k', None)
                if e0 is None or k is None:
                    return
                ehi = ktoe(self.wids['elo'].GetValue()) + e0
                elo = ktoe(self.wids['ehi'].GetValue()) + e0
            self.wids['elo'].SetValue(elo)
            self.wids['ehi'].SetValue(ehi)
            self.fit_xspace = 'e'
            self.wids['fitspace_label'].SetLabel('Fit Range (eV):')


    def show_subframe(self, name, frameclass, **opts):
        shown = False
        if name in self.subframes:
            try:
                self.subframes[name].Raise()
                shown = True
            except:
                del self.subframes[name]
        if not shown:
            self.subframes[name] = frameclass(self, **opts)

    def onPanelExposed(self, **kws):
        # called when notebook is selected
        fname = self.controller.filelist.GetStringSelection()
        if fname in self.controller.file_groups:
            gname = self.controller.file_groups[fname]
            dgroup = self.controller.get_group(gname)
            self.fill_form(dgroup)
            self.process(dgroup=dgroup)

    def write_message(self, msg, panel=0):
        self.controller.write_message(msg, panel=panel)

    def larch_eval(self, cmd):
        """eval"""
        self.command_hist.append(cmd)
        return self.controller.larch.eval(cmd)

    def _plain_larch_eval(self, cmd):
        return self.controller.larch._larch.eval(cmd)

    def get_session_history(self):
        """return full session history"""
        larch = self.controller.larch
        return getattr(larch.input, 'hist_buff',
                       getattr(larch.parent, 'hist_buff', []))

    def larch_get(self, sym):
        """get value from larch symbol table"""
        return self.controller.larch.symtable.get_symbol(sym)

    def build_display(self):
        """build display"""

        self.panel.Add(SimpleText(self.panel, self.title, **titleopts),
                       dcol=7)
        self.panel.Add(SimpleText(self.panel, ' coming soon....'),
                       dcol=7, newrow=True)
        self.panel.pack()

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.panel, 1, wx.LEFT|wx.CENTER, 3)
        pack(self, sizer)

    def set_defaultconfig(self, config):
        """set the default configuration for this session"""
        if self.configname not in self.controller.conf_group:
            self.controller.conf_group[self.configname] = {}
        self.controller.conf_group[self.configname].update(config)

    def get_defaultconfig(self):
        """get the default configuration for this session"""
        return deepcopy(self.controller.get_config(self.configname))

    def get_config(self, dgroup=None, with_erange=True):
        """get and set processing configuration for a group"""
        if dgroup is None:
            dgroup = self.controller.get_group()
        if not hasattr(dgroup, 'config'):
            dgroup.config = Group(__name__='xas_viewer config')
        conf = getattr(dgroup.config, self.configname, None)
        if conf is None:
            conf = self.get_defaultconfig()
            setattr(dgroup.config, self.configname, conf)
        if dgroup is not None and with_erange:
            _emin = min(dgroup.energy)
            _emax = max(dgroup.energy)
            e0 = 5*int(dgroup.e0/5.0)
            conf['elo'] = min(_emax, max(_emin, conf['elo_rel'] + e0))
            conf['ehi'] = min(_emax, max(_emin, conf['ehi_rel'] + e0))
        return conf

    def update_config(self, config, dgroup=None):
        """set/update processing configuration for a group"""
        if dgroup is None:
            dgroup = self.controller.get_group()
        conf = getattr(dgroup.config, self.configname, None)
        if conf is None:
            conf = self.get_defaultconfig()

        conf.update(config)
        if dgroup is not None:
            setattr(dgroup.config, self.configname, conf)

    def fill_form(self, dat):
        if isinstance(dat, Group):
            dat = group2dict(dat)

        for name, wid in self.wids.items():
            if isinstance(wid, FloatCtrl) and name in dat:
                wid.SetValue(dat[name])

    def get_energy_ranges(self, dgroup):
        pass

    def read_form(self):
        "read for, returning dict of values"
        dgroup = self.controller.get_group()
        form_opts = {'groupname': getattr(dgroup, 'groupname', 'No Group')}
        for name, wid in self.wids.items():
            val = None
            for method in ('GetValue', 'GetStringSelection', 'IsChecked',
                           'GetLabel'):
                meth = getattr(wid, method, None)
                if callable(meth):
                    try:
                        val = meth()
                    except TypeError:
                        pass
                if val is not None:
                    break
            form_opts[name] = val
        return form_opts

    def process(self, dgroup=None, **kws):
        """override to handle data process step"""
        if self.skip_process:
            return
        self.skip_process = True

    def add_text(self, text, dcol=1, newrow=True):
        self.panel.Add(SimpleText(self.panel, text),
                       dcol=dcol, newrow=newrow)


    def add_floatspin(self, name, value, with_pin=True,
                      relative_e0=False, **kws):
        """create FloatSpin with Pin button for onSelPoint"""
        if with_pin:
            pin_action = partial(self.xasmain.onSelPoint, opt=name,
                                 relative_e0=relative_e0,
                                 callback=self.pin_callback)
            fspin, pinb = FloatSpinWithPin(self.panel, value=value,
                                           pin_action=pin_action, **kws)
        else:
            fspin = FloatSpin(self.panel, value=value, **kws)
            pinb = None

        self.wids[name] = fspin

        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(fspin)
        if pinb is not None:
            sizer.Add(pinb)
        return sizer

    def pin_callback(self, opt='__', xsel=None, relative_e0=False, **kws):
        """called to do reprocessing after a point is selected as from Pin/Plot"""
        if xsel is not None and opt in self.wids:
            if relative_e0 and 'e0' in self.wids:
                xsel -= self.wids['e0'].GetValue()
            self.wids[opt].SetValue(xsel)
            wx.CallAfter(self.onProcess)

    def onPlot(self, evt=None):
        pass

    def onPlotOne(self, evt=None, dgroup=None, **kws):
        pass

    def onPlotSel(self, evt=None, groups=None, **kws):
        pass

    def onProcess(self, evt=None, **kws):
        pass
